#include <cassert>
#include <cfloat>
#include <cmath>
#include <vector>
#include "maxmin.h"

// Compute a sum of norms of m*n matrix A, n vector x, and m vector b.
static double
norms(const int m,
      const int n,
      const double* A,
      const double* x,
      const double* b)
{
    double anorm=0;
    for(int i=0; i<m; ++i){
        double t=0;
        for(int j=0; j<n; ++j) t+=std::fabs(A[n*i+j]);
        if(t>anorm) anorm=t;
    }
    double xnorm=0;
    for(int i=0; i<n; ++i) xnorm+=std::fabs(x[i]);
    double bnorm=0;
    for(int i=0; i<m; ++i) bnorm+=std::fabs(b[i]);
    return anorm*xnorm+bnorm;
}

// Compute r=b+A*x, and return the value of the minimum entry.
static double
multiply_and_add(const int m,
                 const int n,
                 const double* A,
                 const double* x,
                 const double* b,
                 std::vector<double>& r)
{
    double minr=DBL_MAX;
    for(int i=0; i<m; ++i){
        r[i]=b[i];
        for(int j=0; j<n; ++j)
            r[i]+=A[n*i+j]*x[j];
        if(r[i]<minr) minr=r[i];
    }
    return minr;
}

// Compute A=L*Q for k rows of an m*n matrix A stored row-by-row,
// producing L k*k and Q k*n (assuming space has already been allocated
// for L and Q). Returns false if the factorization shows A to be
// rank-deficient (in this case, the diagonal of L is set to 1 arbitrarily,
// but the corresponding row of Q is left zero.)
static bool
factor_lq(const int m,
          const int n,
          const int k,
          const int* rows,
          const double* A,
          double *L,
          double *Q)
{
    bool full_rank=true;
    // Use Gram-Schmidt on the rows for now - Householder would be better
    for(int i=0; i<k; ++i){ // do row i
        for(int j=0; j<n; ++j) Q[n*i+j]=A[n*rows[i]+j];
        for(int j=0; j<i; ++j){ // project out row j of Q
            double d=0;
            for(int p=0; p<n; ++p) d+=Q[n*j+p]*Q[n*i+p];
            L[k*i+j]=d;
            for(int p=0; p<n; ++p) Q[n*i+p]-=d*Q[n*j+p];
        }
        double rownorm2=0;
        for(int j=0; j<n; ++j) rownorm2+=Q[n*i+j]*Q[n*i+j];
        if(rownorm2){
            double rownorm=std::sqrt(rownorm2);
            L[k*i+i]=rownorm;
            for(int j=0; j<n; ++j) Q[n*i+j]/=rownorm;
        }else{
            L[k*i+i]=1;
            full_rank=false;
        }
    }
    return full_rank;
}

// Solve L*Q*d=1 where L is k*k and Q is k*n, k<=n, stored row-by-row
static void
solve_lq1(const int k,
          const int n,
          const double* L,
          const double* Q,
          std::vector<double>& d,
          double* temp)
{
    for(int i=0; i<k; ++i){
        double t=1;
        for(int j=0; j<i; ++j) t-=L[k*i+j]*temp[j];
        temp[i]=t/L[k*i+i];
    }
    for(int j=0; j<n; ++j){
        double t=0;
        for(int i=0; i<k; ++i) t+=Q[n*i+j]*temp[i];
        d[j]=t;
    }
}

// Compute dot-product of row i of m*n matrix A and n vector d.
static double
dot(const int m,
    const int n,
    const double* A,
    int i,
    const std::vector<double>& d)
{
    double r=0;
    for(int j=0; j<n; ++j)
        r+=A[n*i+j]*d[j];
    return r;
}

// Given an "active" subset of the rows of an m*n matrix A, try to compute a
// direction n vector that has positive dot-products with those rows.
// Return false if it's certain there is no such direction. Returning true
// may not indicate success --- check the direction in that case!
static bool
find_positive_direction(const int m,
                        const int n,
                        const double* A,
                        const std::vector<int>& active,
                        std::vector<double>& d)
{
    std::vector<double> work(m*n+m*m+m);
    double *L=&work[0], *Q=&work[m*m], *temp=&work[m*m+m*n];
    if((int)active.size()<=n){ // underdetermined or balanced: the usual case
        factor_lq(m, n, active.size(), &active[0], A, L, Q);
        solve_lq1(active.size(), n, L, Q, d, temp); 
        return true; // it probably is fine!
    }else{ // overdetermined, possibly --- this is a degenerate case!
        // we have to find, in this case, the best n rows to solve with
        std::vector<int> test(n);
        for(int i=0; i<n; ++i) test[i]=active[i]; // first n as initial guess
        factor_lq(m, n, n, &test[0], A, L, Q);
        solve_lq1(n, n, L, Q, d, temp);
        for(unsigned int iteration=0; iteration<active.size(); ++iteration){
            double worst_d=1; // any positive number would work
            int index=-1;
            for(unsigned int i=0; i<active.size(); ++i){
                double this_d=dot(m, n, A, active[i], d);
                if(this_d<worst_d){
                    worst_d=this_d;
                    index=active[i];
                }
            }
            if(worst_d>0) return true; // we found a good direction!
            // otherwise try sticking index into test
            bool found_good_one=false;
            for(int i=0; i<n; ++i){
                std::swap(test[i], index);
                factor_lq(m, n, n, &test[0], A, L, Q);
                solve_lq1(n, n, L, Q, d, temp);
                if(dot(m, n, A, index, d)>0){ // works for the row we took out?
                    found_good_one=true;
                    break;
                }
                // put the old vector back into test if not
                std::swap(test[i], index);
            }
            if(!found_good_one) return false; // no progress possible
        }
        return true; // we quit early - it might be OK
    }
}

int
solve_maxmin(const int m,
             const int n,
             const double* A,
             const double* b,
             const int use_initial_guess,
             double* x)
{
    assert(n>0);
    assert(A && b && x);
    if(m<=n) return 0;
    int i;
    if(!use_initial_guess)
        for(i=0; i<n; ++i) x[i]=0;
    std::vector<double> r(m, 0), d(n, 0);
    std::vector<int> active(m);
    for(int iteration=0; iteration<m*n; ++iteration){
        // evaluate where we are
        double minr=multiply_and_add(m, n, A, x, b, r)
                   +1e-12*norms(m, n, A, x, b); // to allow for rounding error
        // find all the rows that are at the minimum
        active.resize(0);
        for(i=0; i<m; ++i) if(r[i]<=minr) active.push_back(i);
        // try to find a direction to search on
        if(!find_positive_direction(m, n, A, active, d))
            return 1; // we're done: cannot make any progress
        // among the minimum rows, find the slowest increasing on this direction
        double slowdot=DBL_MAX;
        for(unsigned int p=0; p<active.size(); ++p){
            double r=dot(m, n, A, active[p], d);
            if(r<slowdot) slowdot=r;
        }
        if(slowdot<=0) return 1; // if we can make no more progress, done!
        // find the maximum step size we can take (i.e. where the slow row
        // ends up equal to another row, with positive step size)
        double step=DBL_MAX;
        for(i=0; i<m; ++i){
            double a=dot(m, n, A, i, d);
            if(a>=slowdot) continue; // row i won't become a new minimum!
            double s=(r[i]-minr)/(slowdot-a);
            if(s<step) step=s;
        }
        if(step==DBL_MAX) // unbounded?
            return 0;
        // take the step!
        for(i=0; i<n; ++i) x[i]+=step*d[i];
    }
    return 1;
}
